from GramAddict.core.handle_sources import interact
from GramAddict.core.filter import Filter
import logging
from functools import partial
from colorama import Style
from os import path
from random import shuffle
from GramAddict.core.decorators import run_safely
from GramAddict.core.plugin_loader import Plugin
from GramAddict.core.views import TabBarView
from GramAddict.core.utils import (
    get_value,
)
from GramAddict.core.interaction import (
    _on_interaction,
    interact_with_user,
    is_follow_limit_reached_for_source,
)

logger = logging.getLogger(__name__)


class InteractUsernames(Plugin):
    """Interact with users that are given from a file"""

    def __init__(self):
        super().__init__()
        self.description = "Interact with users that are given from a file"
        self.arguments = [
            {
                "arg": "--aainteract-from-file",
                "nargs": "+",
                "help": "filenames of the list of users [*.txt]",
                "metavar": ("filename1", "filename2"),
                "default": None,
                "operation": True,
            }
        ]

    def run(self, device, configs, storage, sessions, plugin):
        class State:
            def __init__(self):
                pass

            is_job_completed = False

        self.args = configs.args
        self.device_id = configs.args.device
        self.sessions = sessions
        self.session_state = sessions[-1]
        profile_filter = Filter(storage)
        self.current_mode = plugin

        file_list = [file for file in (self.args.interact_from_file)]
        shuffle(file_list)

        for file in file_list:
            limit_reached = self.session_state.check_limit(
                self.args, limit_type=self.session_state.Limit.ALL
            )

            self.state = State()
            logger.info(f"Handle {file}", extra={"color": f"{Style.BRIGHT}"})

            on_interaction = partial(
                _on_interaction,
                likes_limit=int(self.args.total_likes_limit),
                source=file,
                interactions_limit=get_value(
                    self.args.interactions_count, "Interactions count: {}", 70
                ),
                sessions=self.sessions,
                session_state=self.session_state,
                args=self.args,
            )

            def get_int_from_percentage_ranges(
                stories_range, follow_range, comment_range, interact_range
            ):
                stories_percentage = get_value(
                    stories_range, "Chance of watching stories: {}%", 40
                )
                follow_percentage = get_value(
                    follow_range, "Chance of following: {}%", 40
                )
                comment_percentage = get_value(
                    comment_range, "Chance of commenting: {}%", 40
                )
                interact_percentage = get_value(
                    interact_range, "Chance of interacting: {}%", 40
                )
                return (
                    stories_percentage,
                    follow_percentage,
                    comment_percentage,
                    interact_percentage,
                )

            (
                stories_percentage,
                follow_percentage,
                comment_percentage,
                interact_percentage,
            ) = get_int_from_percentage_ranges(
                self.args.stories_percentage,
                self.args.follow_percentage,
                self.args.comment_percentage,
                self.args.interact_percentage,
            )

            @run_safely(
                device=device,
                device_id=self.device_id,
                sessions=self.sessions,
                session_state=self.session_state,
                screen_record=self.args.screen_record,
                configs=configs,
            )
            def job():
                self.handle_username_file(
                    device,
                    file,
                    self.args.likes_count,
                    self.args.stories_count,
                    stories_percentage,
                    follow_percentage,
                    int(self.args.follow_limit) if self.args.follow_limit else None,
                    comment_percentage,
                    interact_percentage,
                    self.args.scrape_to_file,
                    plugin,
                    storage,
                    profile_filter,
                    on_interaction,
                )
                self.state.is_job_completed = True

            while not self.state.is_job_completed and not limit_reached:
                job()

            if limit_reached:
                logger.info("Likes and follows limit reached.")
                self.session_state.check_limit(
                    self.args, limit_type=self.session_state.Limit.ALL, output=True
                )
                break

    def handle_username_file(
        self,
        device,
        current_file,
        likes_count,
        stories_percentage,
        follow_percentage,
        follow_limit,
        comment_percentage,
        interact_percentage,
        scraping_file,
        current_job,
        storage,
        profile_filter,
        on_like,
        on_watch,
        on_interaction,
    ):
        interaction = partial(
            interact_with_user,
            my_username=self.session_state.my_username,
            likes_count=likes_count,
            stories_percentage=stories_percentage,
            follow_percentage=follow_percentage,
            comment_percentage=comment_percentage,
            on_like=on_like,
            on_watch=on_watch,
            profile_filter=profile_filter,
            args=self.args,
            session_state=self.session_state,
            scraping_file=scraping_file,
            current_mode=self.current_mode,
        )

        is_follow_limit_reached = partial(
            is_follow_limit_reached_for_source,
            follow_limit=follow_limit,
            source=current_file,
            session_state=self.session_state,
        )

        need_to_refresh = True
        if path.isfile(current_file):
            with open(current_file, "r") as f:
                for line in f:
                    username = line.strip()
                    if username != "":
                        if storage.is_user_in_blacklist(username):
                            logger.info(f"@{username} is in blacklist. Skip.")
                            continue
                        elif storage.check_user_was_interacted(username):
                            logger.info(f"@{username}: already interacted. Skip.")
                            continue
                        if need_to_refresh:
                            search_view = TabBarView(device).navigateToSearch()
                        profile_view = search_view.navigateToUsername(
                            username, True, need_to_refresh
                        )
                        need_to_refresh = False
                        if not profile_view:
                            continue

                        if not interact(
                            storage,
                            is_follow_limit_reached,
                            username,
                            interaction,
                            device,
                            self.session_state,
                            current_job,
                            on_interaction,
                        ):
                            return
                        device.back()
                    else:
                        logger.info("Line in file is blank, skip.")
                remaining = f.readlines()
            if self.args.delete_interacted_users:
                with open(current_file, "w", encoding="UTF-8") as f:
                    f.writelines(remaining)
        else:
            logger.warning(f"File {current_file} not found.")
            return

        logger.info(f"Interact with users in {current_file} complete.")
        device.back()



tested_ig_version = "185.0.1.39.120"
tested_ig_version_splitted = tested_ig_version.split(".")
running_ig_version = "185.0.1.39.116"
print(f"Instagram version: {running_ig_version}")
running_ig_version_splitted = running_ig_version.split(".")

for n in range(len(running_ig_version_splitted)):
    if int(running_ig_version_splitted[n]) > int(tested_ig_version_splitted[n]):
        print(
            f"You have a newer version of IG then the one we tested! (Tested version: {running_ig_version})"
        )
        break
    else:
        if int(running_ig_version_splitted[n]) == int(tested_ig_version_splitted[n]):
            continue
        break


from GramAddict.core.device_facade import Timeout
import logging

from colorama import Fore
from GramAddict.core.views import (
    AccountView,
    HashTagView,
    LanguageView,
    OptionsView,
    PlacesView,
    PostsGridView,
    ProfileView,
    SettingsView,
    TabBarView,
    UniversalActions,
)

logger = logging.getLogger(__name__)


def check_if_english(device):
    logger.debug("Navigate to PROFILE.")
    ProfileView(device)._click_on_avatar()
    logger.debug("Checking if English..")
    post, follower, following = ProfileView(device)._getSomeText()
    if None not in {post, follower, following}:
        if post == "Posts" and follower == "Followers" and following == "Following":
            logger.debug("Instagram in English..")
        else:
            logger.info("Switching to English locale", extra={"color": f"{Fore.GREEN}"})
            ProfileView(device).navigateToOptions()
            OptionsView(device).navigateToSettings()
            SettingsView(device).navigateToAccount()
            AccountView(device).navigateToLanguage()
            LanguageView(device).setLanguage("english")
            logger.debug(
                "After changing language, IG goes to feed. Let's go to profile view again."
            )
            ProfileView(device)._click_on_avatar()
    else:
        logger.warning(
            "Failed to check your Instagram language. Be sure to set it to English or the bot won't work!"
        )
    return ProfileView(device, is_own_profile=True)


def nav_to_blogger(device, username, current_job):
    """navigate to blogger (followers list or posts)"""
    _to_followers = True if current_job.endswith("followers") else False
    _to_followwing = True if current_job.endswith("following") else False
    if username is None:
        logger.info("Open your followers")
        profile_view = TabBarView(device).navigateToProfile()
        if _to_followers:
            profile_view.navigateToFollowers()
        elif _to_followwing:
            profile_view.navigateToFollowing()
    else:
        search_view = TabBarView(device).navigateToSearch()
        profile_view = search_view.navigateToUsername(username)
        if not profile_view:
            return False

        logger.info(f"Open @{username} followers")
        if _to_followers:
            profile_view.navigateToFollowers()
        elif _to_followwing:
            profile_view.navigateToFollowing()
    return True


def nav_to_hashtag_or_place(device, target, current_job):
    """navigate to hashtag/place/feed list"""
    search_view = TabBarView(device).navigateToSearch()
    if (
        not search_view.navigateToHashtag(target)
        if current_job.startswith("hashtag")
        else not search_view.navigateToPlaces(target)
    ):
        return False

    TargetView = HashTagView if current_job.startswith("hashtag") else PlacesView

    if current_job.endswith("recent"):
        logger.info("Switching to Recent tab.")
        recent_tab = TargetView(device)._getRecentTab()
        if recent_tab.exists(Timeout.MEDIUM):
            recent_tab.click()
        else:
            return False

        if UniversalActions(device)._check_if_no_posts():
            UniversalActions(device)._reload_page()
            if UniversalActions(device)._check_if_no_posts():
                return False

    logger.info("Opening the first result.")

    result_view = TargetView(device)._getRecyclerView()
    TargetView(device)._getFistImageView(result_view).click()
    return True


def nav_to_post_likers(device, username, my_username):
    """navigate to blogger post likers"""
    if username == my_username:
        TabBarView(device).navigateToProfile()
    else:
        search_view = TabBarView(device).navigateToSearch()
        if not search_view.navigateToUsername(username):
            return False
    profile_view = ProfileView(device)
    is_private = profile_view.isPrivateAccount()
    posts_count = profile_view.getPostsCount()
    is_empty = posts_count == 0
    if is_private or is_empty:
        private_empty = "Private" if is_private else "Empty"
        logger.info(f"{private_empty} account.", extra={"color": f"{Fore.GREEN}"})
        return False
    logger.info("Opening the first post")
    ProfileView(device).swipe_to_fit_posts()
    PostsGridView(device).navigateToPost(0, 0)
    return True


def nav_to_feed(device):
    TabBarView(device).navigateToHome()



from GramAddict.core.device_facade import Location, SleepTime, Timeout
from GramAddict.core.session_state import SessionState
from GramAddict.core import storage
import logging
import emoji
from datetime import datetime
from random import randint, shuffle, choice
from typing import Tuple
from time import sleep, time
from os import path
from colorama import Fore, Style
from GramAddict.core.report import print_scrape_report, print_short_report
from GramAddict.core.resources import ClassName, ResourceID as resources
from GramAddict.core.utils import (
    get_value,
    random_sleep,
    append_to_file,
    save_crash,
)
from GramAddict.core.views import (
    MediaType,
    ProfileView,
    CurrentStoryView,
    PostsGridView,
    SearchView,
    UniversalActions,
    Direction,
    case_insensitive_re,
)

logger = logging.getLogger(__name__)

FOLLOW_REGEX = "^Follow$"
FOLLOWBACK_REGEX = "^Follow Back$"
UNFOLLOW_REGEX = "^Following|^Requested"


def load_config(config):
    global args
    global configs
    global ResourceID
    args = config.args
    configs = config
    ResourceID = resources(config.args.app_id)


def interact_with_user(
    device,
    username,
    my_username,
    likes_count,
    stories_percentage,
    can_follow,
    follow_percentage,
    comment_percentage,
    pm_percentage,
    profile_filter,
    args,
    session_state,
    scraping_file,
    current_mode,
) -> Tuple[bool, bool, bool, bool, int, int, int]:
    """
    :return: (whether interaction succeed, whether @username was followed during the interaction, if you scraped that account, if you sent a PM, number of liked, number of watched, number of commented)
    """
    number_of_liked = 0
    number_of_watched = 0
    number_of_commented = 0
    like_succeed = comment_done = interacted = followed = scraped = sent_pm = False

    if username == my_username:
        logger.info("It's you, skip.")
        return (
            interacted,
            followed,
            scraped,
            sent_pm,
            number_of_liked,
            number_of_watched,
            number_of_commented,
        )

    logger.debug("Checking profile..")
    start_time = time()
    profile_data, skipped = profile_filter.check_profile(device, username)
    if skipped:
        delta = format(time() - start_time, ".2f")
        logger.debug(f"Profile checked in {delta}s")
        return (
            interacted,
            followed,
            scraped,
            sent_pm,
            number_of_liked,
            number_of_watched,
            number_of_commented,
        )

    profile_view = ProfileView(device)
    delta = format(time() - start_time, ".2f")
    logger.debug(f"Profile checked in {delta}s")
    if profile_data.is_private or (profile_data.posts_count == 0):
        private_empty = "Private" if profile_data.is_private else "Empty"
        logger.info(f"{private_empty} account.", extra={"color": f"{Fore.GREEN}"})
        if (
            can_send_PM(session_state, pm_percentage)
            and profile_filter.can_pm_to_private_or_empty
        ):
            sent_pm = _send_PM(device, session_state, my_username, 0, True)
        if can_follow and profile_filter.can_follow_private_or_empty():
            if scraping_file is None:
                followed = _follow(
                    device, username, follow_percentage, args, session_state, 0
                )
                return (
                    interacted,
                    followed,
                    scraped,
                    sent_pm,
                    number_of_liked,
                    number_of_watched,
                    number_of_commented,
                )
        else:
            logger.info("Skip user.", extra={"color": f"{Fore.GREEN}"})
            return (
                interacted,
                followed,
                scraped,
                sent_pm,
                number_of_liked,
                number_of_watched,
                number_of_commented,
            )

    # handle the scraping mode
    if scraping_file is not None:
        append_to_file(scraping_file, username)
        logger.info(
            f"Added @{username} at {scraping_file}",
            extra={"color": f"{Style.BRIGHT}{Fore.GREEN}"},
        )
        scraped = True
        return (
            interacted,
            followed,
            scraped,
            sent_pm,
            number_of_liked,
            number_of_watched,
            number_of_commented,
        )

    # if not in scarping mode, we will interact
    number_of_watched = _watch_stories(
        device,
        profile_view,
        username,
        stories_percentage,
        args,
        session_state,
    )

    swipe_amount = ProfileView(device).swipe_to_fit_posts()
    if swipe_amount == -1:
        return (
            interacted,
            followed,
            scraped,
            sent_pm,
            number_of_liked,
            number_of_watched,
            number_of_commented,
        )

    likes_value = get_value(likes_count, "Likes count: {}", 2)
    (
        _,
        _,
        can_comment_job,
    ) = profile_filter.can_comment(current_mode)
    if can_comment_job and comment_percentage != 0:
        max_comments_pro_user = get_value(
            args.max_comments_pro_user, "Max comment count: {}", 1
        )
    if likes_value > 12:
        logger.error("Max number of likes per user is 12.")
        likes_value = 12

    start_time = time()
    full_rows, columns_last_row = profile_view.count_photo_in_view()
    end_time = format(time() - start_time, ".2f")
    photos_indices = list(range(0, full_rows * 3 + (columns_last_row)))

    logger.info(
        f"There {f'is {len(photos_indices)} post' if len(photos_indices)<=1 else f'are {len(photos_indices)} posts'} fully visible. Calculated in {end_time}s"
    )
    if current_mode in [
        "hashtag-posts-recent",
        "hashtag-posts-top",
        "place-posts-recent",
        "place-posts-top",
        "feed",
    ]:
        # in these jobs we did a like already at the post
        photos_indices = photos_indices[1:]
        # sometimes we liked not the last picture, have to introduce the already liked thing..

    if likes_value > len(photos_indices):
        logger.info(
            f"Only {len(photos_indices)} {'photo' if len(photos_indices)<=1 else 'photos'} available."
        )
    else:
        shuffle(photos_indices)
        photos_indices = photos_indices[:likes_value]
        photos_indices = sorted(photos_indices)

    for i in range(0, len(photos_indices)):
        photo_index = photos_indices[i]
        row = photo_index // 3
        column = photo_index - row * 3
        logger.info(f"Open post #{i + 1} ({row + 1} row, {column + 1} column).")
        opened_post_view, media_type, obj_count = PostsGridView(device).navigateToPost(
            row, column
        )

        like_succeed = False
        if opened_post_view:
            _browse_carousel(device, media_type, obj_count)
            like_succeed = do_like(opened_post_view, device, session_state, media_type)
            if like_succeed is True:
                number_of_liked += 1
            if comment_percentage != 0:
                if can_comment(media_type, profile_filter, current_mode):
                    if number_of_commented < max_comments_pro_user:
                        comment_done = _comment(
                            device,
                            my_username,
                            comment_percentage,
                            args,
                            session_state,
                            media_type,
                        )
                        if comment_done:
                            number_of_commented += 1
                    else:
                        logger.info(
                            f"You've already did {max_comments_pro_user} {'comment' if max_comments_pro_user<=1 else 'comments'} for this user!"
                        )
            logger.info("Back to profile.")
            device.back()
        if like_succeed or comment_done:
            interacted = True
        else:
            interacted = False

        if not opened_post_view or not like_succeed:
            reason = "open" if not opened_post_view else "like"
            logger.info(
                f"Could not {reason} photo. Posts count: {profile_data.posts_count}"
            )

            if can_follow and profile_filter.can_follow_private_or_empty():
                followed = _follow(
                    device,
                    username,
                    follow_percentage,
                    args,
                    session_state,
                    swipe_amount,
                )
            else:
                followed = False

            if not followed:
                logger.info("Skip user.", extra={"color": f"{Fore.GREEN}"})
            return (
                interacted,
                followed,
                scraped,
                sent_pm,
                number_of_liked,
                number_of_watched,
                number_of_commented,
            )

    if can_send_PM(session_state, pm_percentage):
        sent_pm = _send_PM(device, session_state, my_username, swipe_amount)
        swipe_amount = 0
    if can_follow:
        return (
            interacted,
            _follow(
                device, username, follow_percentage, args, session_state, swipe_amount
            ),
            scraped,
            sent_pm,
            number_of_liked,
            number_of_watched,
            number_of_commented,
        )

    return (
        interacted,
        followed,
        scraped,
        sent_pm,
        number_of_liked,
        number_of_watched,
        number_of_commented,
    )


def can_send_PM(session_state, pm_percentage):
    pm_chance = randint(1, 100)
    if not session_state.check_limit(
        args, limit_type=session_state.Limit.PM, output=True
    ) and (pm_chance < pm_percentage):
        return True
    else:
        return False


def can_comment(media_type, profile_filter, current_mode):
    (
        can_comment_photos,
        can_comment_videos,
        can_comment_job,
    ) = profile_filter.can_comment(current_mode)
    if can_comment_job:
        if media_type == MediaType.PHOTO and can_comment_photos:
            return True
        elif media_type == MediaType.VIDEO and can_comment_videos:
            return True
    else:
        logger.debug(
            f"Can't comment because filter for {current_mode} in json is: {profile_filter.can_comment(current_mode)}"
        )
        return False


def do_like(opened_post_view, device, session_state, media_type):
    if (
        media_type == MediaType.VIDEO or media_type == MediaType.IGTV
    ) and args.watch_video_time != "0":
        watching_time = get_value(args.watch_video_time, "Watching video for {}s.", 0)
        sleep(watching_time)
    logger.info("Double click post.")
    like_succeed = opened_post_view.likePost()
    if not like_succeed:
        logger.debug("Double click failed. Try the like button.")
        like_succeed = opened_post_view.likePost(click_btn_like=True)

    if like_succeed:
        logger.debug("Like succeed.")
        UniversalActions.detect_block(device)
        session_state.totalLikes += 1
    else:
        logger.warning("Fail to like post. Let's continue...")

    return like_succeed


def is_follow_limit_reached_for_source(session_state, follow_limit, source):
    if follow_limit is None:
        return False

    followed_count = session_state.totalFollowed.get(source)
    return followed_count is not None and followed_count >= follow_limit


def _on_interaction(
    source,
    succeed,
    followed,
    scraped,
    interactions_limit,
    likes_limit,
    sessions,
    session_state,
    args,
):
    session_state = sessions[-1]
    session_state.add_interaction(source, succeed, followed, scraped)

    can_continue = True

    inside_working_hours, _ = SessionState.inside_working_hours(
        args.working_hours, args.time_delta_session
    )
    if not inside_working_hours:
        can_continue = False
    else:
        successful_interactions_count = session_state.successfulInteractions.get(source)
        if (
            successful_interactions_count
            and successful_interactions_count >= interactions_limit
        ):
            logger.info(
                "Reached interaction limit for that source, going to the next one..",
                extra={"color": f"{Fore.CYAN}"},
            )
            can_continue = False

        if args.scrape_to_file is not None:
            if session_state.check_limit(
                args, limit_type=session_state.Limit.SCRAPED, output=False
            ):
                logger.info(
                    "Reached scraped limit, finish.", extra={"color": f"{Fore.CYAN}"}
                )
                can_continue = False
        else:
            if session_state.check_limit(
                args, limit_type=session_state.Limit.LIKES, output=False
            ):
                logger.info(
                    "Reached likes limit, finish.", extra={"color": f"{Fore.CYAN}"}
                )
                can_continue = False

            if session_state.check_limit(
                args, limit_type=session_state.Limit.FOLLOWS, output=False
            ):
                logger.info(
                    "Reached followed limit, finish.", extra={"color": f"{Fore.CYAN}"}
                )
                can_continue = False

            if session_state.check_limit(
                args, limit_type=session_state.Limit.TOTAL, output=False
            ):
                logger.info(
                    "Reached total interaction limit, finish.",
                    extra={"color": f"{Fore.CYAN}"},
                )
                can_continue = False
            if session_state.check_limit(
                args, limit_type=session_state.Limit.SUCCESS, output=False
            ):
                logger.info(
                    "Reached total succesfully interaction limit, finish.",
                    extra={"color": f"{Fore.CYAN}"},
                )
                can_continue = False

    if (can_continue and succeed) or scraped:
        if scraped:
            print_scrape_report(source, session_state)
        else:
            print_short_report(source, session_state)

    return can_continue


def _browse_carousel(device, media_type, obj_count):
    if media_type == MediaType.CAROUSEL:
        carousel_percentage = get_value(configs.args.carousel_percentage, None, 0)
        carousel_count = get_value(configs.args.carousel_count, None, 1)
        if carousel_percentage > randint(0, 100) and carousel_count > 1:
            logger.info("Watching photos/videos in carousel.")
            if obj_count < carousel_count:
                logger.info(f"There are only {obj_count} media in this carousel!")
                carousel_count = obj_count
            n = 1
            while n < carousel_count:
                UniversalActions(device)._swipe_points(
                    direction=Direction.LEFT,
                )
                n += 1


def _comment(device, my_username, comment_percentage, args, session_state, media_type):
    if not session_state.check_limit(
        args, limit_type=session_state.Limit.COMMENTS, output=False
    ):
        comment_chance = randint(1, 100)
        if comment_chance > comment_percentage:
            return False
        # we have to do a little swipe for preventing get the previus post comments button (which is covered by top bar, but present in hierarchy!!)
        UniversalActions(device)._swipe_points(
            direction=Direction.DOWN, delta_y=randint(150, 250)
        )
        tab_bar = device.find(
            resourceId=ResourceID.TAB_BAR,
        )
        media = device.find(
            resourceIdMatches=ResourceID.CAROUSEL_MEDIA_GROUP_AND_ZOOMABLE_VIEW_CONTAINER,
        )
        if int(tab_bar.get_bounds()["top"]) - int(media.get_bounds()["bottom"]) < 150:
            UniversalActions(device)._swipe_points(
                direction=Direction.DOWN, delta_y=randint(150, 250)
            )
        # look at hashtag of comment
        for _ in range(2):
            comment_button = device.find(
                resourceId=ResourceID.ROW_FEED_BUTTON_COMMENT,
            )
            if comment_button.exists():
                logger.info("Open comments of post.")
                comment_button.click()
                comment_box = device.find(
                    resourceId=ResourceID.LAYOUT_COMMENT_THREAD_EDITTEXT,
                    enabled="true",
                )
                if comment_box.exists():
                    comment = load_random_comment(my_username, media_type)
                    if comment is None:
                        SearchView(device)._close_keyboard()
                        device.back()
                        return False
                    logger.info(
                        f"Write comment: {comment}", extra={"color": f"{Fore.CYAN}"}
                    )
                    comment_box.set_text(comment)

                    post_button = device.find(
                        resourceId=ResourceID.LAYOUT_COMMENT_THREAD_POST_BUTTON_CLICK_AREA
                    )
                    post_button.click()
                else:
                    logger.info("Comments on this post have been limited.")
                    SearchView(device)._close_keyboard()
                    device.back()
                    return False

                UniversalActions.detect_block(device)
                SearchView(device)._close_keyboard()
                posted_text = device.find(
                    resourceId=ResourceID.ROW_COMMENT_TEXTVIEW_COMMENT,
                    text=f"{my_username} {comment}",
                )
                when_posted = posted_text.sibling(
                    resourceId=ResourceID.ROW_COMMENT_SUB_ITEMS_BAR
                ).child(resourceId=ResourceID.ROW_COMMENT_TEXTVIEW_TIME_AGO)
                if posted_text.exists(Timeout.MEDIUM) and when_posted.exists(
                    Timeout.MEDIUM
                ):
                    logger.info("Comment succeed.", extra={"color": f"{Fore.GREEN}"})
                    session_state.totalComments += 1
                    comment_confirmed = True
                else:
                    logger.warning("Failed to check if comment succeed.")
                    comment_confirmed = False

                logger.info("Go back to post view.")
                device.back()
                return comment_confirmed
            else:
                like_button = device.find(
                    resourceId=ResourceID.ROW_FEED_BUTTON_LIKE,
                )
                if like_button.exists():
                    logger.info("This post have comments disabled.")
                    return False
                UniversalActions(device)._swipe_points(
                    direction=Direction.DOWN, delta_y=randint(150, 250)
                )
                continue
    return False


def _send_PM(device, session_state, my_username, swipe_amount, private_or_empty=False):
    if private_or_empty:
        options = device.find(
            classNameMatches=ClassName.FRAME_LAYOUT,
            descriptionMatches=case_insensitive_re("^Options$"),
        )
        if options.exists(Timeout.SHORT):
            options.click()
        else:
            return False
        send_pm = device.find(
            classNameMatches=ClassName.BUTTON,
            textMatches=case_insensitive_re("^Send Message$"),
        )
        if send_pm.exists(Timeout.SHORT):
            send_pm.click()
        else:
            return False
    else:
        coordinator_layout = device.find(resourceId=ResourceID.COORDINATOR_ROOT_LAYOUT)
        if coordinator_layout.exists() and swipe_amount != 0:
            UniversalActions(device)._swipe_points(
                direction=Direction.UP, delta_y=swipe_amount
            )
        message_button = device.find(
            classNameMatches=ClassName.BUTTON, clickable=True, textMatches="Message"
        )
        if message_button.exists(Timeout.SHORT):
            message_button.click()
        else:
            return False
    message_box = device.find(
        resourceId=ResourceID.ROW_THREAD_COMPOSER_EDITTEXT,
        className=ClassName.EDIT_TEXT,
        enabled="true",
    )
    if message_box.exists():
        message = load_random_message(my_username)
        if message is None:
            logger.warning("You forgot to populate your PM list!")
            device.back()
            return False
        logger.info(
            f"Write private message: {message}", extra={"color": f"{Fore.CYAN}"}
        )
        message_box.set_text(message)
        send_button = device.find(
            resourceId=ResourceID.ROW_THREAD_COMPOSER_BUTTON_SEND,
            className=ClassName.TEXT_VIEW,
        )
        send_button.click()
        UniversalActions.detect_block(device)
        SearchView(device)._close_keyboard()
        posted_text = device.find(
            resourceId=ResourceID.DIRECT_TEXT_MESSAGE_TEXT_VIEW,
            text=f"{message}",
        )
        message_sending_icon = device.find(
            resourceId=ResourceID.ACTION_ICON, className=ClassName.IMAGE_VIEW
        )
        if message_sending_icon.exists():
            random_sleep()
        if posted_text.exists(Timeout.MEDIUM) and not message_sending_icon.exists():
            logger.info("PM send succeed.", extra={"color": f"{Fore.GREEN}"})
            session_state.totalPm += 1
            pm_confirmed = True
        else:
            logger.warning("Failed to check if PM send succeed.")
            pm_confirmed = False
        logger.info("Go back to profile view.")
        device.back()
        return pm_confirmed
    else:
        logger.info("PM to this user have been limited.")
        SearchView(device)._close_keyboard()
        device.back()
        return False


def load_random_message(my_username):
    def nonblank_lines(f):
        for ln in f:
            line = ln.rstrip()
            if line:
                yield line

    lines = []
    file_name = f"{storage.ACCOUNTS}/{my_username}/{storage.FILENAME_MESSAGES}"
    if path.isfile(file_name):
        with open(file_name, "r", encoding="utf-8") as f:
            for line in nonblank_lines(f):
                lines.append(line)
            random_message = choice(lines)
            if random_message != "":
                return emoji.emojize(random_message, use_aliases=True)
            else:
                return None


def load_random_comment(my_username, media_type):
    def nonblank_lines(f):
        for ln in f:
            line = ln.rstrip()
            if line:
                yield line

    lines = []
    file_name = f"{storage.ACCOUNTS}/{my_username}/{storage.FILENAME_COMMENTS}"
    if path.isfile(file_name):
        with open(file_name, "r", encoding="utf-8") as f:
            for line in nonblank_lines(f):
                lines.append(line)
            try:
                photo_header = lines.index("%PHOTO")
                video_header = lines.index("%VIDEO")
                carousel_header = lines.index("%CAROUSEL")
            except:
                logger.warning(
                    f"You didn't follow the rules of sections for {file_name}! Look at config example."
                )
                return None
            photo_comments = lines[photo_header + 1 : video_header]
            video_comments = lines[video_header + 1 : carousel_header]
            carousel_comments = lines[carousel_header + 1 :]
            if media_type == MediaType.PHOTO:
                random_comment = (
                    choice(photo_comments) if len(photo_comments) > 0 else ""
                )
            if media_type == MediaType.VIDEO or media_type == MediaType.IGTV:
                random_comment = (
                    choice(video_comments) if len(video_comments) > 0 else ""
                )
            if media_type == MediaType.CAROUSEL:
                random_comment = (
                    choice(carousel_comments) if len(carousel_comments) > 0 else ""
                )
            if random_comment != "":
                return emoji.emojize(random_comment, use_aliases=True)
            else:
                return None
    else:
        logger.warning(f"{file_name} not found!")
        return None


def _follow(device, username, follow_percentage, args, session_state, swipe_amount):
    if not session_state.check_limit(
        args, limit_type=session_state.Limit.FOLLOWS, output=False
    ):
        follow_chance = randint(1, 100)
        if follow_chance > follow_percentage:
            return False

        coordinator_layout = device.find(resourceId=ResourceID.COORDINATOR_ROOT_LAYOUT)
        if coordinator_layout.exists(Timeout.MEDIUM) and swipe_amount != 0:
            UniversalActions(device)._swipe_points(
                direction=Direction.UP, delta_y=swipe_amount
            )

        follow_button = device.find(
            clickable=True,
            textMatches=FOLLOW_REGEX,
        )

        if not follow_button.exists():
            unfollow_button = device.find(
                clickable=True,
                textMatches=UNFOLLOW_REGEX,
            )
            followback_button = device.find(
                clickable=True,
                textMatches=FOLLOWBACK_REGEX,
            )
            if unfollow_button.exists():
                logger.info(
                    f"You already follow @{username}.", extra={"color": f"{Fore.GREEN}"}
                )
                return False
            elif followback_button.exists():
                logger.info(
                    f"@{username} already follows you.",
                    extra={"color": f"{Fore.GREEN}"},
                )
                return False
            else:
                logger.error(
                    "Cannot find neither Follow button, Follow Back button, nor Unfollow button."
                )
                save_crash(device)

        follow_button.click()
        if device.find(
            clickable=True,
            textMatches=UNFOLLOW_REGEX,
        ).exists(Timeout.SHORT):
            logger.info(f"Followed @{username}", extra={"color": f"{Fore.GREEN}"})
            UniversalActions.detect_block(device)
            return True
        else:
            logger.info(
                f"Looks like I was not able to follow @{username}",
                extra={"color": f"{Fore.RED}"},
            )
            UniversalActions.detect_block(device)
            return False
    else:
        logger.info("Reached total follows limit, not following.")
        return False


def _watch_stories(
    device,
    profile_view,
    username,
    stories_percentage,
    args,
    session_state,
):
    if not session_state.check_limit(
        args, limit_type=session_state.Limit.WATCHES, output=False
    ):
        story_chance = randint(1, 100)
        if story_chance > stories_percentage:
            return 0

        stories_to_watch = get_value(args.stories_count, "Stories count: {}.", 0)

        if stories_to_watch > 6:
            logger.error("Max number of stories per user is 6.")
            stories_to_watch = 6

        if stories_to_watch == 0:
            return 0
        stories_ring = profile_view.StoryRing()
        if stories_ring.exists():
            stories_counter = 0
            logger.debug("Open the first story.")
            stories_ring.click(sleep=SleepTime.ZERO)
            story_view = CurrentStoryView(device)
            random_sleep(1, 2, modulable=False, logging=False)
            story_view.getStoryFrame().wait(Timeout.SHORT)

            if profile_view.getUsername(error=False) != username:
                start = datetime.now()
                session_state.totalWatched += 1
                stories_counter += 1
                for _ in range(0, 7):
                    random_sleep(0.5, 1, modulable=False, logging=False)
                    if profile_view.getUsername(error=False) == username:
                        break

                if stories_to_watch > 1:
                    story_view = CurrentStoryView(device)
                    for _iter in range(0, stories_to_watch - 1):
                        if story_view.getUsername() == username:
                            try:
                                story_frame = story_view.getStoryFrame()
                                if (
                                    story_frame.exists()
                                    and _iter <= stories_to_watch - 1
                                ):
                                    logger.debug("Going to the next story..")
                                    story_frame.click(
                                        mode=Location.RIGHTEDGE,
                                        sleep=SleepTime.ZERO,
                                    )
                                    session_state.totalWatched += 1
                                    stories_counter += 1
                                    for _ in range(0, 7):
                                        random_sleep(
                                            0.5, 1, modulable=False, logging=False
                                        )
                                        if (
                                            profile_view.getUsername(error=False)
                                            == username
                                        ):
                                            break

                            except Exception as e:
                                logger.error(e)
                                save_crash(device)
                                break
                        else:
                            break

                for _ in range(0, 4):
                    if profile_view.getUsername(error=False) != username:
                        device.back()
                    else:
                        break
                logger.info(
                    f"Watched stories for {(datetime.now()-start).total_seconds():.2f}s."
                )
                return stories_counter
            else:
                logger.warning("Failed to open the story container.")
                return False
        return 0
    else:
        logger.info("Reached total watch limit, not watching stories.")
        return False
